/* 
 * scout_thread.C
 *
 * x-kernel v3.3
 *
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 */

/*
 * Threads package.
 */
#include <strings.h>

#define NPRIORITIES	(SCOUT_THREAD_PRIORITY_MIN \
			 - SCOUT_THREAD_PRIORITY_MAX + 1)

struct scout_stack {
    union {
	void			*sp;		/* top of stack */
	struct scout_stack	*next;		/* next free stack */
    } u;
    void	*stack_limit;
    long	magic;				/* magic number */
};


int trace_scout_thread = 0;

static void start (void)  XNORETURN;	/* forward declaration */
static void retire (void) XNORETURN;	/* never returns */
static void sched (void);

static Scout_Thread_Options def_opts =
{
    "noname",
    0,
    DEFAULT_STACK_SIZE,
    SCOUT_THREAD_PRIORITY_STD
};
const Scout_Thread_Options	*scout_thread_default_options = &def_opts;

/*
 * Invariant: the currently running thread is "scout_thread_self" (it is
 * *not* in any ready_q!).
 */
Scout_Thread			*scout_thread_self = 0;

static Scout_Thread	*dead_man = 0;
static Scout_Thread	*idle_thread = 0;
static u_int		high_prio = SCOUT_THREAD_PRIORITY_MAX;
static Scout_Queue	ready_q[NPRIORITIES];

static struct scout_stack *first_stack = 0;


#define ATTACH_STACK(s) \
{ \
    (s) = first_stack; \
    if (s PREDICT_TRUE) { \
	first_stack = first_stack->u.next; \
    } else { \
	(s) = mkstack(); \
    } /* if */ \
    (s)->u.sp = (s)->stack_limit; \
}

#define DETACH_STACK(s) \
{ \
    (s)->u.next = first_stack; \
    first_stack = (s); \
}


#ifdef SCOUT_DEBUG
#  define STACK_CHECK() \
{ \
    if (scout_thread_free_stack() <= 0 || \
	scout_thread_self->stack->magic != MAGIC_COOKIE) \
    { \
	error("stack overflow!"); \
	retire(); \
    } /* if */ \
}
#else
# define STACK_CHECK()
#endif /* SCOUT_DEBUG */


#ifdef SCOUT_DEBUG

#define MAX_THREADS	256

static Scout_Thread	*blocked[MAX_THREADS];

#define HASH(a)	(((u_long)(a) >> 3) % MAX_THREADS)

/*
 * Print string using minimal amount of stack space:
 */
static void
error(const char *msg)
{
    int ch;

    while ((ch = *msg++) != '\0') {
	putchar(ch);
    } /* while */
    putchar('\n');
} /* error */

#endif /* SCOUT_DEBUG */


static struct scout_stack*
mkstack(void)
{
    struct scout_stack *s;

    /*
     * For now, we have only one size of stacks (i.e., we ignore stack
     * size option).
     */
    s = xMalloc(DEFAULT_STACK_SIZE);
    if (!s PREDICT_FALSE) {
	Kabort(__FILE__": mkstack: out of memory");
    } /* if */
    s->stack_limit = (char*)s + DEFAULT_STACK_SIZE;
    s->magic = MAGIC_COOKIE;
    return s;
} /* mkstack */


/*
 * Instead of an endless loop calling scout_yield(), use
 * continuations.  This avoids having to waste a stack for the idle
 * thread.
 */
static void
idler(void *arg)
{
    scout_queue_append(&ready_q[scout_thread_self->option.priority],
		       scout_thread_self);
    scout_thread_self->running = FALSE;
    DETACH_STACK(scout_thread_self->stack);
    scout_thread_self->stack = 0;
    scout_thread_self = 0;		/* no need to save any state */
    sched();
} /* idler */


/*
 * Switch context from scout_thread_self to t.  This is tricky code.  You
 * should verify the generated assembly code.
 */
static void
cswtch(Scout_Thread *t)
{
    static Scout_Thread *old;

    old = scout_thread_self;
    scout_thread_self = t;
    if (old) {
	SAVE_STATE(old->stack->u.sp);
    } /* if */
    if (scout_thread_self->running) {
	LOAD_STATE(scout_thread_self->stack->u.sp);
    } else {
	scout_thread_self->running = TRUE;
	/*
	 * Switch to new stack and invoke start(); the second argument
	 * is just there to tell GCC that start() is really used.
	 */
	SWITCH_STACK_AND_START(scout_thread_self->stack->u.sp, start);
    } /* if */
} /* cswtch */


static void
print_ready_queue (void)
{
  struct scout_queue_link *el;
  int i;

  printf ("Printing ready queue:\n");
  for (i = 0; i <= 2; i++) {
    printf ("Priority %d: ", i);
    for (el = ready_q[i].head; el; el = el->next) {
      printf ("%p ", el);
    }
    printf ("\n");
  }
  printf ("\n");
}

static void
sched(void)
{
    Scout_Thread *next;
    
/*    print_ready_queue(); */

    /* dispatch to highest priority thread: */
    while (1) {
	assert(high_prio <= SCOUT_THREAD_PRIORITY_MIN);
	next = (Scout_Thread*) scout_queue_remove(&ready_q[high_prio]);
	if (next) {
	    if (!next->stack) {
		ATTACH_STACK(next->stack);
	    } /* if */
	    cswtch(next);
	    if (dead_man) {
		ScoutTrace(scout_thread, TR_EVENTS,
			   (__FILE__": "
			    "freeing space of dead thread `%s' at %p",
			    scout_thread_get_name(dead_man), dead_man));
		DETACH_STACK(dead_man->stack);
		xFree(dead_man);
		dead_man = 0;
	    } /* if */
	    return;
	} /* if */
	++high_prio;
    } /* while */
} /* sched */


static Scout_Thread*
create(Scout_Thread_Func fun, const Scout_Thread_Options *options)
{
    Scout_Thread *t;
    char *mem;
    int name_len;

    name_len = strlen(options->name) + 1;

    /*
     * Make sure requested stack-size is reasonable:
     */
    if (options->stack_size < MIN_USEFUL_STACK_SIZE) {
	ScoutTrace(scout_thread, TR_ERRORS,
		   (__FILE__": "
		    "stack too small (%d bytes, should be >%d bytes)",
		    options->stack_size, MIN_USEFUL_STACK_SIZE));
	return 0;
    } /* if */
    if (options->priority > SCOUT_THREAD_PRIORITY_MIN) {
	ScoutTrace(scout_thread, TR_ERRORS,
		   (__FILE__": illegal thread priority %d specified",
		    options->priority));
	return 0;
    } /* if */

    /*
     * Allocate thread descriptor, and name string in one chunk:
     */
    mem = malloc(sizeof(Scout_Thread) + name_len);
    if (!mem) {
	ScoutTrace(scout_thread, TR_EVENTS,
		   (__FILE__": not enough memory for new thread"));
	return 0;
    } /* if */

    /* initialize memory: */
    t = (Scout_Thread*) mem;
    t->option		= *options;
    t->option.name	= (char*) t + sizeof(Scout_Thread);
    t->running		= FALSE;
    t->fun		= fun;
    t->stack		= 0;
    strcpy(t->option.name, options->name);

    return t;
} /* create */


/*
 * Retire the currently executing thread and invoke the scheduler.
 * All memory associated with the current thread is freed.
 */
static void
retire(void)
{
    dead_man = scout_thread_self;	/* let next thread deallocate space */
    scout_thread_self = 0;		/* avoid state saving */
    sched();
} /* retire */


/*
 * Invoke the initial function of thread scout_thread_self.  If it ever
 * returns, we simply retire the thread.
 */
static void
start(void)
{
    if (dead_man) {
	ScoutTrace(scout_thread, TR_EVENTS,
		   (__FILE__":start: freeing space of dead thread `%s' at %p",
		    scout_thread_get_name(dead_man), dead_man));
	DETACH_STACK(dead_man->stack);
	xFree(dead_man);
	dead_man = 0;
    } /* if */

    ScoutTrace(scout_thread, TR_EVENTS,
	       (__FILE__": starting thread `%s' at %p free stack %d bytes",
		scout_thread_get_name(scout_thread_self),
		scout_thread_self, scout_thread_free_stack()));

    (*scout_thread_self->fun)(scout_thread_self->option.arg);

    ScoutTrace(scout_thread, TR_EVENTS,
	       (__FILE__": thread `%s' (%#p) terminating.",
		scout_thread_get_name(scout_thread_self), scout_thread_self));
    retire();
} /* start */


void
scout_thread_init(Scout_Thread_Func user_init, void *user_arg)
{
    int prio;
    Scout_Thread_Options to;
    Scout_Thread *t;

    for (prio = 0; prio < NPRIORITIES; ++prio) {
	scout_queue_init(&ready_q[prio]);
    } /* for */

    /* create initial thread: */
    to = def_opts;
    to.name = "initializer";
    to.arg  = user_arg;
    t = create(user_init, &to);
    if (!t) {
	Kabort("scout_thread_init: failed to create thread `%s'",
		    to.name);
    } /* if */
    scout_thread_wakeup(t);

    /* create idle thread: */
    to = def_opts;
    to.name = "idler";
    to.arg  = 0;
    to.priority = SCOUT_THREAD_PRIORITY_MIN;
    idle_thread = create(idler, &to);
    if (!idle_thread) {
	Kabort("scout_thread_init: failed to create thread `%s'",
		    to.name);
    } /* if */
    scout_thread_wakeup(idle_thread);

    /* terminate boot stack and get things going: */
    retire();
} /* scout_thread_init */


Scout_Thread*
scout_thread_create(Scout_Thread_Func fun, const Scout_Thread_Options *options)
{
    Scout_Thread *p;
    STACK_CHECK();
    p = create(fun, options);
    return p;
} /* scout_thread_create */


void
scout_thread_stop(void)
{
    ScoutTrace(scout_thread, TR_EVENTS,
	       ("scout_thread_stop: thread `%s' (%p) terminating.",
		scout_thread_get_name(scout_thread_self), scout_thread_self));
    retire();
} /* Scout_ThreadStop */


void
scout_thread_suspend(void)
{
    STACK_CHECK();
    sched();
} /* scout_thread_suspend */


void
scout_thread_suspend_with_continuation(void)
{
    STACK_CHECK();
    scout_thread_self->running = FALSE;
    DETACH_STACK(scout_thread_self->stack);
    scout_thread_self->stack = 0;
    scout_thread_self = 0;		/* no need to save any state */
    sched();
} /* scout_thread_suspend_with_continuation */


void
scout_thread_wakeup(Scout_Thread *t)
{
    assert(t->option.priority <= SCOUT_THREAD_PRIORITY_MIN);
    scout_queue_append(&ready_q[t->option.priority], t);
    if (t->option.priority < high_prio) {
	high_prio = t->option.priority;
    } /* if */
} /* scout_thread_wakeup */


void
scout_thread_yield(void)
{
    STACK_CHECK();
    scout_queue_append(&ready_q[scout_thread_self->option.priority],
		       scout_thread_self);
    sched();
} /* scout_thread_yield */


int
scout_thread_free_stack(void)
{
    char sp;

    return (&sp - (char*)scout_thread_self->stack) - MIN_STACK_SIZE;
} /* scout_thread_free_stack */

			/*** end of scout_thread.c ***/
