/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: mcount.c,v $
 *
 * HISTORY
 * $Log: mcount.c,v $
 * Revision 1.4  1996/02/02  15:53:16  slm
 * Updated copyright and version.
 *
 * Revision 1.3  1995/03/09  18:11:42  davidm
 * (mcount): protect against recursive invocation (on the Alpha, integer
 * division results in function calls that are profiled).
 *
 * Revision 1.2  1994/11/21  18:16:17  davidm
 * Renamed GMON_PROF_* to SCOUT_MON_*.
 *
 * Revision 1.1  1994/11/20  06:40:44  davidm
 * Initial revision
 */
/*-
 * Copyright (c) 1983, 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if !defined(lint) && !defined(KERNEL) && defined(LIBC_SCCS)
/* from: static char sccsid[] = "@(#)mcount.c	8.1 (Berkeley) 6/4/93"; */
static char *rcsid = "$Id: mcount.c,v 1.4 1996/02/02 15:53:16 slm Exp $";
#endif

#include <sys/types.h>
#include "scout_monitor.h"
#include "scout_sync_arch.h"
#include "gmon.h"

#ifdef MCOUNT_CHECK_STACK
# include "scout_thread.h"
#endif

/*
 * mcount is called on entry to each function compiled with the profiling
 * switch set.  _mcount(), which is declared in a machine-dependent way
 * with _MCOUNT_DECL, does the actual work and is either inlined into a
 * C routine or called by an assembly stub.
 *
 * _mcount updates data structures that represent traversals of the
 * program's call graph edges.  frompc and selfpc are the return
 * address and function address that represents the given call graph edge.
 * 
 * Note: the original BSD code used the same variable (frompcindex) for
 * both frompcindex and frompc.  Any reasonable, modern compiler will
 * perform this optimization.
 */
#ifdef MCOUNT_TRACE_CALLS
typedef (*Func)(void);

int tracing_enabled = 1;
struct trace_record {
    Func from, to;
} tracebuf[64*1024];
u_int tracenext = 0;
#endif

void
mcount(u_long frompc, u_long selfpc)
{
#if defined(MCOUNT_CHECK_STACK)
    static int busy = 0;
    long remaining;

    if (busy || !scout_thread_self || !scout_thread_self->stack) {
	return;
    } /* if */
    busy = 1;

    remaining = scout_thread_free_stack();
    if (remaining < 2048) {
	printf("[mcount]: %lx -> %lx\n", frompc, selfpc);
	printf("[mcount]: %ld bytes left on stack (%#lx->%#lx)\n",
	       remaining, frompc, selfpc);
    } /* if */

    busy = 0;
#elif defined(MCOUNT_TRACE_CALLS)
    if (tracing_enabled) {
	tracebuf[tracenext].from = (Func) frompc;
	tracebuf[tracenext].to   = (Func) selfpc;
	if (++tracenext >= sizeof(tracebuf) / sizeof(tracebuf[0])) {
	    tracenext = 0;
	} /* if */;
    } /* if */
#else
    u_short *frompcindex;
    struct tostruct *top, *prevtop;
    struct gmonparam *p;
    long toindex;
    Scout_IPL old_ipl;
    static int active = 0;
    
    p = &_gmonparam;
    /*
     * check that we are profiling
     * and that we aren't recursively invoked.
     */
    if (!(p->state & SCOUT_MON_CNT)) {
	return;
    } /* if */
    
    old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);

    if (active) {
	return;		/* avoid endless recursion */
    }
    active = 1;

    /*
     * check that frompcindex is a reasonable pc value.
     * for example:	signal catchers get called from the stack,
     *		not from text space.  too bad.
     */
    frompc -= p->lowpc;
    if (frompc > p->textsize)
      goto _done;
    
    frompcindex = &p->froms[frompc / (p->hashfraction*sizeof(*p->froms))];
    toindex = *frompcindex;
    if (toindex == 0) {
	/*
	 *	first time traversing this arc
	 */
	toindex = ++p->tos[0].link;
	if (toindex >= p->tolimit)
	  /* halt further profiling */
	  goto _overflow;

	*frompcindex = toindex;
	top = &p->tos[toindex];
	top->selfpc = selfpc;
	top->count = 1;
	top->link = 0;
	goto _done;
    }
    top = &p->tos[toindex];
    if (top->selfpc == selfpc) {
	/*
	 * arc at front of chain; usual case.
	 */
	top->count++;
	goto _done;
    }
    /*
     * have to go looking down chain for it.
     * top points to what we are looking at,
     * prevtop points to previous top.
     * we know it is not at the head of the chain.
     */
    for (;			/* goto _done */; ) {
	if (top->link == 0) {
	    /*
	     * top is end of the chain and none of the chain
	     * had top->selfpc == selfpc.
	     * so we allocate a new tostruct
	     * and link it to the head of the chain.
	     */
	    toindex = ++p->tos[0].link;
	    if (toindex >= p->tolimit)
	      goto _overflow;

	    top = &p->tos[toindex];
	    top->selfpc = selfpc;
	    top->count = 1;
	    top->link = *frompcindex;
	    *frompcindex = toindex;
	    goto _done;
	}
	/*
	 * otherwise, check the next arc on the chain.
	 */
	prevtop = top;
	top = &p->tos[top->link];
	if (top->selfpc == selfpc) {
	    /*
	     * there it is.
	     * increment its count
	     * move it to the head of the chain.
	     */
	    top->count++;
	    toindex = prevtop->link;
	    prevtop->link = top->link;
	    top->link = *frompcindex;
	    *frompcindex = toindex;
	    goto _done;
	}
		
    }
_done:
    active = 0;
    scout_swap_ipl(old_ipl);
    return;

_overflow:
    p->overflow = TRUE;
    goto _done;
#endif
} /* mcount */

			/*** end of mcount.c ***/
